library fields.model;

import 'dart:async';
import 'dart:collection';
import 'dart:mirrors';
import 'package:tags/tags.dart';

part 'src/model/default_editors.dart';
part 'src/model/field.dart';
part 'src/model/field_editor.dart';
part 'src/model/field_list.dart';
part 'src/model/meta.dart';
part 'src/model/mirrors.dart';
part 'src/model/value_transform.dart';

void registerDefaultFieldEditors() {  
  if (!FieldEditor.isDefaultRegistered)
    FieldEditor.registerDefault(DefaultEditor);  
  if (!FieldEditor.isRegistered(bool)) 
    FieldEditor.register(bool, BoolEditor);
  if (!FieldEditor.isRegistered(DateTime))
    FieldEditor.register(DateTime, DateTimeEditor);
  if (!FieldEditor.isRegistered(num))
    FieldEditor.register(num, NumberEditor);
}

void registerDefaultValueTransforms() {  
  ValueTransform.register(const ValueTransform<double, num>(_double2num));
  ValueTransform.register(const ValueTransform<int, num>(_int2num));
  ValueTransform.register(const ValueTransform<num, double>(_num2double));
  ValueTransform.register(const ValueTransform<num, int>(_num2int));
  ValueTransform.register(const ValueTransform<String, int>(_string2int));
  ValueTransform.register(const ValueTransform<String, double>(_string2double));
}

/// A list of types that are atomic from the viewpoint of this library and its
/// reflection.
/// 
/// Some user-defined types may lead to a [StackOverflowError] when this library
/// performs reflection of instance members.  For example, when a class declares
/// a getter that allocates a new object of its own class on each call.  This 
/// list provides a way to prevent reflective invocation of such getters.
/// 
/// This list is initialized with a default list that contains simple types from
/// `dart:core`; it may be modified to register user-defined atomic types.
final List<Type> atomicTypes = new List.from(_atomicTypes);
