part of fields.model;

/// Please star <http://dartbug.com/11186>.
typedef bool Predicate<E>(E element);

// TODO(rms): https://code.google.com/p/dart/issues/detail?id=11178
class FieldList extends ListBase<Field> {
  
  static final Comparator defaultComparator = 
      (Field a, Field b) => a.name.compareTo(b.name);
  
  static final Predicate defaultPredicate = (Field f) => true;
  
  final List<Field> _source = new List<Field>();
  final List<Field> _view = new List<Field>();
  
  final Field _parent;
  final _listeners = new List<StreamSubscription>();
  
  Comparator<Field> _comparator = defaultComparator;
  Comparator<Field> 
    get comparator => _comparator;
    set comparator(Comparator<Field> c) {
      assert(c != null);
      if (_comparator != c) {
        _comparator = c;
        for (Field f in _source) f.children.comparator = c;        
        refreshView();
      }
    }
    
  final _fieldChanged = new StreamController<Field>();
  Stream<Field> get fieldChanged => _fieldChanged.stream;
  
  Predicate<Field> _predicate = defaultPredicate;
  Predicate<Field>
    get predicate => _predicate;
    set predicate(Predicate<Field> p) {
      assert(p != null);
      if (_predicate != p) {
        _predicate = p;
        for (Field f in _source) f.children.predicate = p;        
        refreshView();
      }
    }
    
  int 
    get length => _view.length;
    set length(int length) => 
        throw new UnsupportedError('FieldList is not modifiable.');
      
  FieldList([this._parent]);
  
  Field operator [](int index) => _view[index];
  
  void operator []=(int index, Field value) => 
      throw new UnsupportedError('FieldList is not modifiable.');
    
  List<Field> _reflect(Object object) {
    final fields = new List<Field>();    
    final instanceMirror = reflect(object);
    // We reflect the ClassMirror from `Object.runtimeType` instead of getting
    // the ClassMirror from the `InstanceMirror.type` to support proxy objects.
    final classMirror = reflectClass(object.runtimeType);
    classMirror.declarations.forEach((key, declaration) {
      if (declaration is MethodMirror && 
          declaration.isGetter && 
          !declaration.isStatic) {
        final field = new Field.forGetter(
            key,
            instanceMirror, 
            getter: declaration, 
            parent: _parent);
        if (field != null) {
          fields.add(field);
        }        
      } else if (declaration is VariableMirror && !declaration.isStatic) {
        final field = new Field.forVariable(
            key,
            instanceMirror, 
            variable: declaration,
            parent: _parent);
        if (field != null) {
          fields.add(field);
        }
      }
    });
    return fields;
  }
  
  void refreshSource(Object object) {
    final fields = _reflect(object);
    fields.forEach((field) {
      // Here we consider fields equal if their declaredName & type match.
      final predicate = (f) => 
          f.declaredName == field.declaredName && f.type == field.type;
      if (_source.any(predicate)) {
        _source.singleWhere(predicate)
            .._getter = field._getter
            .._setter = field._setter
            ..refreshValue();
      }
      else {
        // Fields need to watch all of their siblings.
        fields
          .where((f) => f != field)
          .forEach((f) => f.values.listen((_) => field.refreshValue()));
        
        _source.add(field);
      }
    });
    refreshView();
  }
      
  void refreshView() {    
    _listeners.forEach((s) => s.cancel());
    _listeners.clear();    
    _view
    ..clear()
    ..addAll(_source.where(predicate).toList()..sort(comparator))
    ..forEach((Field f) {
      _listeners.add(f.values.listen((_) {
        if (_fieldChanged.hasListener) {
          _fieldChanged.add(f);
        }
      }));
      f.children.refreshView();
    });
  }
}
